#lang sicp

;;;;METACIRCULAR EVALUATOR THAT SEPARATES ANALYSIS FROM EXECUTION
;;;; FROM SECTION 4.1.7 OF STRUCTURE AND INTERPRETATION OF COMPUTER PROGRAMS

;; OP: Modified to run in racket.

(#%require racket/trace)
(#%require "syntax.rkt")
(#%require "environment.rkt")
(#%require "common.rkt")

(#%provide analyzing-driver-loop
           analyzing-eval-program)

;;;SECTION 4.1.7

(define (analyzing-eval exp env)
  ((analyze exp) env))

(define (analyze exp)
  (cond ((self-evaluating? exp) 
         (analyze-self-evaluating exp))
        ((quoted? exp) (analyze-quoted exp))
        ((variable? exp) (analyze-variable exp))
        ((assignment? exp) (analyze-assignment exp))
        ((definition? exp) (analyze-definition exp))
        ((if? exp) (analyze-if exp))
        ((lambda? exp) (analyze-lambda exp))
        ((begin? exp) (analyze-sequence (begin-actions exp)))
        ((cond? exp) (analyze (cond->if exp)))
        ((let? exp) (analyze (let->combination exp)))
        ((application? exp) (analyze-application exp))
        (else
         (error "Unknown expression type -- ANALYZE" exp))))

(define (analyze-self-evaluating exp)
  (lambda (env) exp))

(define (analyze-quoted exp)
  (let ((qval (text-of-quotation exp)))
    (lambda (env) qval)))

(define (analyze-variable exp)
  (lambda (env) (lookup-variable-value exp env)))

(define (analyze-assignment exp)
  (let ((var (assignment-variable exp))
        (vproc (analyze (assignment-value exp))))
    (lambda (env)
      (set-variable-value! var (vproc env) env)
      'ok)))

(define (analyze-definition exp)
  (let ((var (definition-variable exp))
        (vproc (analyze (definition-value exp))))
    (lambda (env)
      (define-variable! var (vproc env) env)
      'ok)))

(define (analyze-if exp)
  (let ((pproc (analyze (if-predicate exp)))
        (cproc (analyze (if-consequent exp)))
        (aproc (analyze (if-alternative exp))))
    (lambda (env)
      (if (true? (pproc env))
          (cproc env)
          (aproc env)))))

(define (analyze-lambda exp)
  (let ((vars (lambda-parameters exp))
        (bproc (analyze-sequence (scan-out-defines (lambda-body exp)))))
    (lambda (env) (make-procedure vars bproc env))))

(define (analyze-sequence exps)
  (define (sequentially proc1 proc2)
    (lambda (env) (proc1 env) (proc2 env)))
  (define (loop first-proc rest-procs)
    (if (null? rest-procs)
        first-proc
        (loop (sequentially first-proc (car rest-procs))
              (cdr rest-procs))))
  (let ((procs (map analyze exps)))
    (if (null? procs)
        (error "Empty sequence -- ANALYZE"))
    (loop (car procs) (cdr procs))))

(define (analyze-application exp)
  (let ((fproc (analyze (operator exp)))
        (aprocs (map analyze (operands exp))))
    (lambda (env)
      (execute-application (fproc env)
                           (map (lambda (aproc) (aproc env))
                                aprocs)))))

(define (execute-application proc args)
  (cond ((primitive-procedure? proc)
         (apply-primitive-procedure proc args))
        ((compound-procedure? proc)
         ((procedure-body proc)
          (extend-environment (procedure-parameters proc)
                              args
                              (procedure-environment proc))))
        (else
         (error
          "Unknown procedure type -- EXECUTE-APPLICATION"
          proc))))

(define input-prompt ";;; MA-Eval input:")
(define output-prompt ";;; MA-Eval value:")

(define (analyzing-driver-loop)
  (define env (setup-environment))
  (define (driver-loop)
    (prompt-for-input input-prompt)
    (let ((input (read)))
      (let ((output (analyzing-eval input env)))
        (announce-output output-prompt)
        (user-print output)))
    (driver-loop))
  (driver-loop))

(define (analyzing-eval-program program)
  (define env (setup-environment))
  (eval-program program analyzing-eval
                user-print env))

'ANALYZING-METACIRCULAR-EVALUATOR-LOADED

;; Exercise 4.23: For a sequence containing a single procedure:
;;
;; (In the below, pn is the result of (analyze procn))
;; Text analyser:
;; (analyze-sequence (proc1))
;; -> (loop p1 ())
;; -> p1
;; 
;; Alyssa analyser:
;; (analyze-sequence (proc1))
;; -> (lambda (env) (execute-sequence (p1) env))
;; -> (lambda (env) (p1 env))

;; Here, the second analyser introduces an additional procedure
;; application rather than just returning the result of (analyze
;; proc1). 

;; For two procedures:
;;
;; Text:
;; (analyze-sequence (proc1 proc2))
;; -> (loop p1 (p2))
;; -> (loop (lambda (e1) (p1 e1) (p2 e1)) ())  ;; This procedure
;; application is done during analyze-sequence.
;; -> (lambda (e1) (p1 e1) (p2 e1))
;; 
;; Alyssa analyser:
;; (analyse-sequence (proc1 proc2))
;; -> (lambda (env) (execute-sequence (p1 p2) env))  ;;
;; analyze-sequence returns this procdure (the recursive calls to
;; execute-sequence that follow are not called until the procedure is evaluated).

;; When the procedure is evaluated:
;; -> (lambda (env) (p1 env) (execute-sequence (p2) env))
;; -> (lambda (env) (p1 env) (p2 env))

;; So the result is the same, but the first analyser returns a
;; procedure that returns the result in a single application, whereas
;; the second requires further evaluation to get the result.
