#lang sicp

;;;;LAZY EVALUATOR FROM SECTION 4.2 OF
;;;; STRUCTURE AND INTERPRETATION OF COMPUTER PROGRAMS

;;;; OMP: modified to work with racket and to fit with existing code

;;;; This evaluator uses the modifications of exercise 4.31: arguments
;;;; can be strict, lazy or lazy with memoization.  To specify a
;;;; procedure, strict in x, lazy in y and lazy with memoization in z:
;;;; (define (f x (y 'lazy) (z 'lazy-memo))).  Also modified from the
;;;; book's version is the use of data-directed dispatch for eval.

(#%require racket/trace)
(#%require "syntax.rkt")
(#%require "environment.rkt")
(#%require "common.rkt")
(#%require "special-forms.rkt")
(#%require "dispatch-table.rkt")

(#%provide lazy-driver-loop
           lazy-eval-program
           lazy-put-dispatch!
           actual-value)

(define type-tag car)
(define dispatch-table (make-dispatch-table))

(define lazy-dispatch-alist
  `((if . ,(lambda (exp env)
             (lazy-eval-if exp env)))
    (begin .
           ,(lambda (exp env)
              (eval-sequence (begin-actions exp) env lazy-eval)))
    (set! . ,(lambda (exp env)
               (eval-assignment exp env lazy-eval)))
    (define .
      ,(lambda (exp env)
         (eval-definition exp env lazy-eval)))
    (make-unbound! . ,(lambda (exp env)
                        (eval-unbind exp env lazy-eval)))
    (quote .
           ,(lambda (exp env)
              (text-of-quotation exp)))
    (lambda .
      ,(lambda (exp env)
         (make-procedure (lambda-parameters exp)
                         (scan-out-defines (lambda-body exp))
                         env)))
    (cond .
          ,(lambda (exp env)
             (lazy-eval (cond->if exp) env)))
    (and . ,(lambda (exp env)
              (eval-and exp env lazy-eval)))
    (or . ,(lambda (exp env)
             (eval-or exp env lazy-eval)))
    (let . ,(lambda (exp env)
              (lazy-eval (let->combination exp) env)))
    (let* . ,(lambda (exp env)
               (lazy-eval (let*->nested-let exp) env)))
    (letrec . ,(lambda (exp env)
                 (lazy-eval (letrec->let exp) env)))
    (for . ,(lambda (exp env)
              (lazy-eval (for->named-let exp) env)))
    (unless . ,(lambda (exp env)
                 (lazy-eval (unless->if exp) env)))))

;; Set up dispatch table from the alist
(put-alist! dispatch-table lazy-dispatch-alist)

;; Allow setting of dispatched procedures externally
(define (lazy-put-dispatch! tag proc)
  (put! dispatch-table tag proc))

(define (lazy-eval exp env)
  (cond ((self-evaluating? exp) exp)
        ((variable? exp) (lookup-variable-value exp env))
        ((get dispatch-table (type-tag exp))
         => (lambda (proc) (proc exp env)))
        ((application? exp)             ; clause from book
         (lazy-apply (actual-value (operator exp) env)
                     (operands exp)
                     env))
        (else
         (error "Unknown expression type -- LAZY-EVAL" exp))))

(define (actual-value exp env)
  (force-it (lazy-eval exp env)))

(define (lazy-apply procedure arguments env)
  (cond ((primitive-procedure? procedure)
         (apply-primitive-procedure
          procedure
          (list-of-arg-values arguments env))) ; changed
        ((compound-procedure? procedure)
         (eval-sequence
          (procedure-body procedure)
          (extend-environment
           (procedure-parameters procedure)
           (list-of-qualified-args
            arguments
            (procedure-delays procedure)
            env) ; changed
           (procedure-environment procedure))
          lazy-eval))
        (else
         (error
          "Unknown procedure type -- APPLY" procedure))))

(define (list-of-arg-values exps env)
  (if (no-operands? exps)
      '()
      (cons (actual-value (first-operand exps) env)
            (list-of-arg-values (rest-operands exps)
                                env))))

(define (list-of-qualified-args exps delays env)
  (if (no-operands? exps)
      '()
      (cons
       (let ((delay (first-delay delays)))
         (if (eq? delay 'strict)
           (lazy-eval (first-operand exps) env)
           (delay-it (first-operand exps) env delay)))
       (list-of-qualified-args
        (rest-operands exps)
        (rest-delays delays)
        env))))

(define (lazy-eval-if exp env)
  (if (true? (actual-value (if-predicate exp) env))
      (lazy-eval (if-consequent exp) env)
      (lazy-eval (if-alternative exp) env)))

(define input-prompt ";;; L-Eval input:")
(define output-prompt ";;; L-Eval value:")

(define (lazy-driver-loop)
  (define env (setup-environment))
  (define (driver-loop)
    (prompt-for-input input-prompt)
    (let ((input (read)))
      (let ((output (actual-value input env)))
        (announce-output output-prompt)
        (user-print output))
      (driver-loop)))
  (driver-loop))

(define (lazy-eval-program program)
  (define env (setup-environment))
  (eval-program program actual-value user-print env))

;;; Representing thunks

(define (delay-it exp env method)
  (cond ((eq? method 'lazy-memo)
         (make-memo-thunk exp env))
        ((eq? method 'lazy)
         (make-thunk exp env))
        (else
         (error ("Unknown delay type -- DELAY-IT" method)))))

;; memo-thunks created by are memoized on forcing. thunks are not.
(define (force-it obj)
  (cond ((memo-thunk? obj)
         (let ((result (actual-value
                        (thunk-exp obj)
                        (thunk-env obj))))
           (set-car! obj 'evaluated-thunk)
           (set-car! (cdr obj) result)  ; replace exp with its value
           (set-cdr! (cdr obj) '())     ; forget unneeded env
           result))
        ((evaluated-thunk? obj)
         (thunk-value obj))
        ((thunk? obj)
         (actual-value
          (thunk-exp obj)
          (thunk-env obj)))
        (else obj)))

(#%require (only racket/base module+))
(module+ test
  (#%require rackunit)
  (define env (setup-environment))
  (check-equal? (force-it (delay-it '(+ 1 1) env 'lazy)) 2)
  (check-equal? (force-it (delay-it '(+ 1 1) env 'lazy-memo)) 2))

'LAZY-EVALUATOR-LOADED
